#pragma once

#include <cstdint>
#include <type_traits>
#include "Lambda.hpp"

template <typename Head, typename Cons>
struct List {
	using type = typename Head::type;
	using value = Head;
	using cons = Cons;
};

template <>
struct List<void, void> {
	using type = void;
	using value = void;
};

template <typename Head>
struct List<Head, void> {
	using type = typename Head::type;
	using value = Head;
	using cons = ::List<void, void>;
};

template<typename Head, typename... Cons>
struct ListConstructor {
	using value = ::List<Head, typename ListConstructor<Cons...>::value>;
};

template<typename Head>
struct ListConstructor<Head> {
	using value = ::List<Head, void>;
};

template<typename List, uint64_t N>
struct List_Nth_Element {
	using value = typename List_Nth_Element<typename List::cons, N - 1>::value;
};

template<typename List>
struct List_Nth_Element<List, 0> {
	using value = typename List::value;
};


template<template <typename, typename> typename Func, typename Init, typename List, bool = std::is_same<typename List::type, void>::value>
struct Fold :
	public Lambda<Init, Func<typename List::type, Init>, Init, List>
{
	using value = Init;
};

template<template <typename, typename> typename Func, typename Init, typename List>
struct Fold<Func, Init, List, false> :
	public Lambda<Init, Func<typename List::type, Init>, Init, List>
{
	using value = Func<typename List::value, typename Fold<Func, Init, typename List::cons>::value>::value;
};


template<template <typename> typename Func, typename List, bool = std::is_same<typename List::type, void>::value>
struct Map :
	public Lambda<List, Func<typename List::type>, List>
{
	using value = ::List<void, void>;
};

template<template <typename> typename Func, typename List>
struct Map<Func, List, false> :
	public Lambda<List, Func<typename List::type>, List>
{
	using value = ::List<typename Func<typename List::value>::value, typename Map<Func, typename List::cons>::value>;
};
